"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)
    :codeauthor: Alexandru Bleotu (alexandru.bleotu@morganstanley.com)


    salt.utils.schema
    ~~~~~~~~~~~~~~~~~

    Object Oriented Configuration - JSON Schema compatible generator

    This code was inspired by `jsl`__, "A Python DSL for describing JSON
    schemas".

    .. __: https://jsl.readthedocs.io/


    A configuration document or configuration document section is defined using
    the py:class:`Schema`, the configuration items are defined by any of the
    subclasses of py:class:`BaseSchemaItem` as attributes of a subclass of
    py:class:`Schema` class.

    A more complex configuration document (containing a defininitions section)
    is defined using the py:class:`DefinitionsSchema`. This type of
    schema supports having complex configuration items as attributes (defined
    extending the py:class:`ComplexSchemaItem`). These items have other
    configuration items (complex or not) as attributes, allowing to verify
    more complex JSON data structures

    As an example:

    .. code-block:: python

        class HostConfig(Schema):
            title = 'Host Configuration'
            description = 'This is the host configuration'

            host = StringItem(
                'Host',
                'The looong host description',
                default=None,
                minimum=1
            )

            port = NumberItem(
                description='The port number',
                default=80,
                required=False,
                minimum=0,
                inclusiveMinimum=False,
                maximum=65535
            )

    The serialized version of the above configuration definition is:

    .. code-block:: python

        >>> print(HostConfig.serialize())
        OrderedDict([
            ('$schema', 'http://json-schema.org/draft-04/schema#'),
            ('title', 'Host Configuration'),
            ('description', 'This is the host configuration'),
            ('type', 'object'),
            ('properties', OrderedDict([
                ('host', {'minimum': 1,
                          'type': 'string',
                          'description': 'The looong host description',
                          'title': 'Host'}),
                ('port', {'description': 'The port number',
                          'default': 80,
                          'inclusiveMinimum': False,
                          'maximum': 65535,
                          'minimum': 0,
                          'type': 'number'})
            ])),
            ('required', ['host']),
            ('x-ordering', ['host', 'port']),
            ('additionalProperties', True)]
        )
        >>> print(salt.utils.json.dumps(HostConfig.serialize(), indent=2))
        {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "title": "Host Configuration",
            "description": "This is the host configuration",
            "type": "object",
            "properties": {
                "host": {
                    "minimum": 1,
                    "type": "string",
                    "description": "The looong host description",
                    "title": "Host"
                },
                "port": {
                    "description": "The port number",
                    "default": 80,
                    "inclusiveMinimum": false,
                    "maximum": 65535,
                    "minimum": 0,
                    "type": "number"
                }
            },
            "required": [
                "host"
            ],
            "x-ordering": [
                "host",
                "port"
            ],
            "additionalProperties": false
        }


    The serialized version of the configuration block can be used to validate a
    configuration dictionary using the `python jsonschema library`__.

    .. __: https://pypi.python.org/pypi/jsonschema

    .. code-block:: python

        >>> import jsonschema
        >>> jsonschema.validate({'host': 'localhost', 'port': 80}, HostConfig.serialize())
        >>> jsonschema.validate({'host': 'localhost', 'port': -1}, HostConfig.serialize())
        Traceback (most recent call last):
        File "<stdin>", line 1, in <module>
        File "/usr/lib/python2.7/site-packages/jsonschema/validators.py", line 478, in validate
            cls(schema, *args, **kwargs).validate(instance)
        File "/usr/lib/python2.7/site-packages/jsonschema/validators.py", line 123, in validate
            raise error
        jsonschema.exceptions.ValidationError: -1 is less than the minimum of 0

        Failed validating 'minimum' in schema['properties']['port']:
            {'default': 80,
            'description': 'The port number',
            'inclusiveMinimum': False,
            'maximum': 65535,
            'minimum': 0,
            'type': 'number'}

        On instance['port']:
            -1
        >>>


    A configuration document can even be split into configuration sections. Let's reuse the above
    ``HostConfig`` class and include it in a configuration block:

    .. code-block:: python

        class LoggingConfig(Schema):
            title = 'Logging Configuration'
            description = 'This is the logging configuration'

            log_level = StringItem(
                'Logging Level',
                'The logging level',
                default='debug',
                minimum=1
            )

        class MyConfig(Schema):

            title = 'My Config'
            description = 'This my configuration'

            hostconfig = HostConfig()
            logconfig = LoggingConfig()


    The JSON Schema string version of the above is:

    .. code-block:: python

        >>> print salt.utils.json.dumps(MyConfig.serialize(), indent=4)
        {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "title": "My Config",
            "description": "This my configuration",
            "type": "object",
            "properties": {
                "hostconfig": {
                    "id": "https://non-existing.saltstack.com/schemas/hostconfig.json#",
                    "title": "Host Configuration",
                    "description": "This is the host configuration",
                    "type": "object",
                    "properties": {
                        "host": {
                            "minimum": 1,
                            "type": "string",
                            "description": "The looong host description",
                            "title": "Host"
                        },
                        "port": {
                            "description": "The port number",
                            "default": 80,
                            "inclusiveMinimum": false,
                            "maximum": 65535,
                            "minimum": 0,
                            "type": "number"
                        }
                    },
                    "required": [
                        "host"
                    ],
                    "x-ordering": [
                        "host",
                        "port"
                    ],
                    "additionalProperties": false
                },
                "logconfig": {
                    "id": "https://non-existing.saltstack.com/schemas/logconfig.json#",
                    "title": "Logging Configuration",
                    "description": "This is the logging configuration",
                    "type": "object",
                    "properties": {
                        "log_level": {
                            "default": "debug",
                            "minimum": 1,
                            "type": "string",
                            "description": "The logging level",
                            "title": "Logging Level"
                        }
                    },
                    "required": [
                        "log_level"
                    ],
                    "x-ordering": [
                        "log_level"
                    ],
                    "additionalProperties": false
                }
            },
            "additionalProperties": false
        }

        >>> import jsonschema
        >>> jsonschema.validate(
            {'hostconfig': {'host': 'localhost', 'port': 80},
             'logconfig': {'log_level': 'debug'}},
            MyConfig.serialize())
        >>> jsonschema.validate(
            {'hostconfig': {'host': 'localhost', 'port': -1},
             'logconfig': {'log_level': 'debug'}},
            MyConfig.serialize())
        Traceback (most recent call last):
        File "<stdin>", line 1, in <module>
        File "/usr/lib/python2.7/site-packages/jsonschema/validators.py", line 478, in validate
            cls(schema, *args, **kwargs).validate(instance)
        File "/usr/lib/python2.7/site-packages/jsonschema/validators.py", line 123, in validate
            raise error
        jsonschema.exceptions.ValidationError: -1 is less than the minimum of 0

        Failed validating 'minimum' in schema['properties']['hostconfig']['properties']['port']:
            {'default': 80,
            'description': 'The port number',
            'inclusiveMinimum': False,
            'maximum': 65535,
            'minimum': 0,
            'type': 'number'}

        On instance['hostconfig']['port']:
            -1
        >>>

    If however, you just want to use the configuration blocks for readability
    and do not desire the nested dictionaries serialization, you can pass
    ``flatten=True`` when defining a configuration section as a configuration
    subclass attribute:

    .. code-block:: python

        class MyConfig(Schema):

            title = 'My Config'
            description = 'This my configuration'

            hostconfig = HostConfig(flatten=True)
            logconfig = LoggingConfig(flatten=True)


    The JSON Schema string version of the above is:

    .. code-block:: python

        >>> print(salt.utils.json.dumps(MyConfig, indent=4))
        {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "title": "My Config",
            "description": "This my configuration",
            "type": "object",
            "properties": {
                "host": {
                    "minimum": 1,
                    "type": "string",
                    "description": "The looong host description",
                    "title": "Host"
                },
                "port": {
                    "description": "The port number",
                    "default": 80,
                    "inclusiveMinimum": false,
                    "maximum": 65535,
                    "minimum": 0,
                    "type": "number"
                },
                "log_level": {
                    "default": "debug",
                    "minimum": 1,
                    "type": "string",
                    "description": "The logging level",
                    "title": "Logging Level"
                }
            },
            "x-ordering": [
                "host",
                "port",
                "log_level"
            ],
            "additionalProperties": false
        }
"""

import inspect
import textwrap

import salt.utils.args

# import salt.utils.yaml
from salt.utils.odict import OrderedDict

BASE_SCHEMA_URL = "https://non-existing.saltstack.com/schemas"
RENDER_COMMENT_YAML_MAX_LINE_LENGTH = 80


class NullSentinel:
    """
    A class which instance represents a null value.
    Allows specifying fields with a default value of null.
    """

    def __bool__(self):
        return False

    __nonzero__ = __bool__


Null = NullSentinel()
"""
A special value that can be used to set the default value
of a field to null.
"""


# make sure nobody creates another Null value
def _failing_new(*args, **kwargs):
    raise TypeError("Can't create another NullSentinel instance")


NullSentinel.__new__ = staticmethod(_failing_new)
del _failing_new


class SchemaMeta(type):
    @classmethod
    def __prepare__(mcs, name, bases):
        return OrderedDict()

    def __new__(mcs, name, bases, attrs):
        # Mark the instance as a configuration document/section
        attrs["__config__"] = True
        attrs["__flatten__"] = False
        attrs["__config_name__"] = None

        # Let's record the configuration items/sections
        items = {}
        sections = {}
        order = []
        # items from parent classes
        for base in reversed(bases):
            if hasattr(base, "_items"):
                items.update(base._items)
            if hasattr(base, "_sections"):
                sections.update(base._sections)
            if hasattr(base, "_order"):
                order.extend(base._order)

        # Iterate through attrs to discover items/config sections
        for key, value in attrs.items():
            entry_name = None
            if not hasattr(value, "__item__") and not hasattr(value, "__config__"):
                continue
            if hasattr(value, "__item__"):
                # the value is an item instance
                if hasattr(value, "title") and value.title is None:
                    # It's an item instance without a title, make the title
                    # its name
                    value.title = key
                entry_name = value.__item_name__ or key
                items[entry_name] = value
            if hasattr(value, "__config__"):
                entry_name = value.__config_name__ or key
                sections[entry_name] = value
            order.append(entry_name)

        attrs["_order"] = order
        attrs["_items"] = items
        attrs["_sections"] = sections
        return type.__new__(mcs, name, bases, attrs)

    def __call__(cls, flatten=False, allow_additional_items=False, **kwargs):
        instance = object.__new__(cls)
        instance.__config_name__ = kwargs.pop("name", None)
        if flatten is True:
            # This configuration block is to be treated as a part of the
            # configuration for which it was defined as an attribute, not as
            # its own sub configuration
            instance.__flatten__ = True
        if allow_additional_items is True:
            # The configuration block only accepts the configuration items
            # which are defined on the class. On additional items, validation
            # with jsonschema will fail
            instance.__allow_additional_items__ = True
        instance.__init__(**kwargs)
        return instance


class BaseSchemaItemMeta(type):
    """
    Config item metaclass to "tag" the class as a configuration item
    """

    @classmethod
    def __prepare__(mcs, name, bases):
        return OrderedDict()

    def __new__(mcs, name, bases, attrs):
        # Register the class as an item class
        attrs["__item__"] = True
        attrs["__item_name__"] = None
        # Instantiate an empty list to store the config item attribute names
        attributes = []
        for base in reversed(bases):
            try:
                base_attributes = getattr(base, "_attributes", [])
                if base_attributes:
                    attributes.extend(base_attributes)
                # Extend the attributes with the base argspec argument names
                # but skip "self"
                for argname in salt.utils.args.get_function_argspec(base.__init__).args:
                    if argname == "self" or argname in attributes:
                        continue
                    if argname == "name":
                        continue
                    attributes.append(argname)
            except TypeError:
                # On the base object type, __init__ is just a wrapper which
                # triggers a TypeError when we're trying to find out its
                # argspec
                continue
        attrs["_attributes"] = attributes
        return type.__new__(mcs, name, bases, attrs)

    def __call__(cls, *args, **kwargs):
        # Create the instance class
        instance = object.__new__(cls)
        if args:
            raise RuntimeError(
                "Please pass all arguments as named arguments. Un-named "
                "arguments are not supported"
            )
        for key in kwargs.copy():
            # Store the kwarg keys as the instance attributes for the
            # serialization step
            if key == "name":
                # This is the item name to override the class attribute name
                instance.__item_name__ = kwargs.pop(key)
                continue
            if key not in instance._attributes:
                instance._attributes.append(key)
        # Init the class
        instance.__init__(*args, **kwargs)
        # Validate the instance after initialization
        for base in reversed(inspect.getmro(cls)):
            validate_attributes = getattr(base, "__validate_attributes__", None)
            if validate_attributes:
                if (
                    instance.__validate_attributes__.__func__.__code__
                    is not validate_attributes.__code__
                ):
                    # The method was overridden, run base.__validate_attributes__ function
                    base.__validate_attributes__(instance)
        # Finally, run the instance __validate_attributes__ function
        instance.__validate_attributes__()
        # Return the initialized class
        return instance


class Schema(metaclass=SchemaMeta):
    """
    Configuration definition class
    """

    # Define some class level attributes to make PyLint happier
    title = None
    description = None
    _items = _sections = _order = None
    __flatten__ = False
    __allow_additional_items__ = False

    @classmethod
    def serialize(cls, id_=None):
        # The order matters
        serialized = OrderedDict()
        if id_ is not None:
            # This is meant as a configuration section, sub json schema
            serialized["id"] = "{}/{}.json#".format(BASE_SCHEMA_URL, id_)
        else:
            # Main configuration block, json schema
            serialized["$schema"] = "http://json-schema.org/draft-04/schema#"
        if cls.title is not None:
            serialized["title"] = cls.title
        if cls.description is not None:
            if cls.description == cls.__doc__:
                serialized["description"] = textwrap.dedent(cls.description).strip()
            else:
                serialized["description"] = cls.description

        required = []
        ordering = []
        serialized["type"] = "object"
        properties = OrderedDict()
        cls.after_items_update = []
        for name in cls._order:  # pylint: disable=E1133
            skip_order = False
            item_name = None
            if name in cls._sections:  # pylint: disable=E1135
                section = cls._sections[name]
                serialized_section = section.serialize(
                    None if section.__flatten__ is True else name
                )
                if section.__flatten__ is True:
                    # Flatten the configuration section into the parent
                    # configuration
                    properties.update(serialized_section["properties"])
                    if "x-ordering" in serialized_section:
                        ordering.extend(serialized_section["x-ordering"])
                    if "required" in serialized_section:
                        required.extend(serialized_section["required"])
                    if hasattr(section, "after_items_update"):
                        cls.after_items_update.extend(section.after_items_update)
                    skip_order = True
                else:
                    # Store it as a configuration section
                    properties[name] = serialized_section

            if name in cls._items:  # pylint: disable=E1135
                config = cls._items[name]
                item_name = config.__item_name__ or name
                # Handle the configuration items defined in the class instance
                if config.__flatten__ is True:
                    serialized_config = config.serialize()
                    cls.after_items_update.append(serialized_config)
                    skip_order = True
                else:
                    properties[item_name] = config.serialize()

                if config.required:
                    # If it's a required item, add it to the required list
                    required.append(item_name)

            if skip_order is False:
                # Store the order of the item
                if item_name is not None:
                    if item_name not in ordering:
                        ordering.append(item_name)
                else:
                    if name not in ordering:
                        ordering.append(name)

        if properties:
            serialized["properties"] = properties

        # Update the serialized object with any items to include after properties.
        # Do not overwrite properties already existing in the serialized dict.
        if cls.after_items_update:
            after_items_update = {}
            for entry in cls.after_items_update:
                for name, data in entry.items():
                    if name in after_items_update:
                        if isinstance(after_items_update[name], list):
                            after_items_update[name].extend(data)
                    else:
                        after_items_update[name] = data
            if after_items_update:
                after_items_update.update(serialized)
                serialized = after_items_update

        if required:
            # Only include required if not empty
            serialized["required"] = required
        if ordering:
            # Only include ordering if not empty
            serialized["x-ordering"] = ordering
        serialized["additionalProperties"] = cls.__allow_additional_items__
        return serialized

    @classmethod
    def defaults(cls):
        serialized = cls.serialize()
        defaults = {}
        for name, details in serialized["properties"].items():
            if "default" in details:
                defaults[name] = details["default"]
                continue
            if "properties" in details:
                for sname, sdetails in details["properties"].items():
                    if "default" in sdetails:
                        defaults.setdefault(name, {})[sname] = sdetails["default"]
                continue
        return defaults

    @classmethod
    def as_requirements_item(cls):
        serialized_schema = cls.serialize()
        required = serialized_schema.get("required", [])
        for name in serialized_schema["properties"]:
            if name not in required:
                required.append(name)
        return RequirementsItem(requirements=required)

    # @classmethod
    # def render_as_rst(cls):
    #    '''
    #    Render the configuration block as a restructured text string
    #    '''
    #    # TODO: Implement RST rendering
    #    raise NotImplementedError

    # @classmethod
    # def render_as_yaml(cls):
    #    '''
    #    Render the configuration block as a parseable YAML string including comments
    #    '''
    #    # TODO: Implement YAML rendering
    #    raise NotImplementedError


class SchemaItem(metaclass=BaseSchemaItemMeta):
    """
    Base configuration items class.

    All configurations must subclass it
    """

    # Define some class level attributes to make PyLint happier
    __type__ = None
    __format__ = None
    _attributes = None
    __flatten__ = False

    __serialize_attr_aliases__ = None

    required = False

    def __init__(self, required=None, **extra):
        """
        :param required: If the configuration item is required. Defaults to ``False``.
        """
        if required is not None:
            self.required = required
        self.extra = extra

    def __validate_attributes__(self):
        """
        Run any validation check you need the instance attributes.

        ATTENTION:

        Don't call the parent class when overriding this
        method because it will just duplicate the executions. This class'es
        metaclass will take care of that.
        """
        if self.required not in (True, False):
            raise RuntimeError("'required' can only be True/False")

    def _get_argname_value(self, argname):
        """
        Return the argname value looking up on all possible attributes
        """
        # Let's see if there's a private function to get the value
        argvalue = getattr(self, "__get_{}__".format(argname), None)
        if argvalue is not None and callable(argvalue):
            argvalue = argvalue()  # pylint: disable=not-callable
        if argvalue is None:
            # Let's see if the value is defined as a public class variable
            argvalue = getattr(self, argname, None)
        if argvalue is None:
            # Let's see if it's defined as a private class variable
            argvalue = getattr(self, "__{}__".format(argname), None)
        if argvalue is None:
            # Let's look for it in the extra dictionary
            argvalue = self.extra.get(argname, None)
        return argvalue

    def serialize(self):
        """
        Return a serializable form of the config instance
        """
        raise NotImplementedError


class BaseSchemaItem(SchemaItem):
    """
    Base configuration items class.

    All configurations must subclass it
    """

    # Let's define description as a class attribute, this will allow a custom configuration
    # item to do something like:
    #   class MyCustomConfig(StringItem):
    #       '''
    #       This is my custom config, blah, blah, blah
    #       '''
    #       description = __doc__
    #
    description = None
    # The same for all other base arguments
    title = None
    default = None
    enum = None
    enumNames = None

    def __init__(
        self,
        title=None,
        description=None,
        default=None,
        enum=None,
        enumNames=None,
        **kwargs
    ):
        """
        :param required:
            If the configuration item is required. Defaults to ``False``.
        :param title:
            A short explanation about the purpose of the data described by this item.
        :param description:
            A detailed explanation about the purpose of the data described by this item.
        :param default:
            The default value for this configuration item. May be :data:`.Null` (a special value
            to set the default value to null).
        :param enum:
            A list(list, tuple, set) of valid choices.
        """
        if title is not None:
            self.title = title
        if description is not None:
            self.description = description
        if default is not None:
            self.default = default
        if enum is not None:
            self.enum = enum
        if enumNames is not None:
            self.enumNames = enumNames
        super().__init__(**kwargs)

    def __validate_attributes__(self):
        if self.enum is not None:
            if not isinstance(self.enum, (list, tuple, set)):
                raise RuntimeError(
                    "Only the 'list', 'tuple' and 'set' python types can be used "
                    "to define 'enum'"
                )
            if not isinstance(self.enum, list):
                self.enum = list(self.enum)
        if self.enumNames is not None:
            if not isinstance(self.enumNames, (list, tuple, set)):
                raise RuntimeError(
                    "Only the 'list', 'tuple' and 'set' python types can be used "
                    "to define 'enumNames'"
                )
            if len(self.enum) != len(self.enumNames):
                raise RuntimeError(
                    "The size of 'enumNames' must match the size of 'enum'"
                )
            if not isinstance(self.enumNames, list):
                self.enumNames = list(self.enumNames)

    def serialize(self):
        """
        Return a serializable form of the config instance
        """
        serialized = {"type": self.__type__}
        for argname in self._attributes:
            if argname == "required":
                # This is handled elsewhere
                continue
            argvalue = self._get_argname_value(argname)
            if argvalue is not None:
                if argvalue is Null:
                    argvalue = None
                # None values are not meant to be included in the
                # serialization, since this is not None...
                if (
                    self.__serialize_attr_aliases__
                    and argname in self.__serialize_attr_aliases__
                ):
                    argname = self.__serialize_attr_aliases__[argname]
                serialized[argname] = argvalue
        return serialized

    def __get_description__(self):
        if self.description is not None:
            if self.description == self.__doc__:
                return textwrap.dedent(self.description).strip()
            return self.description

    # def render_as_rst(self, name):
    #    '''
    #    Render the configuration item as a restructured text string
    #    '''
    #    # TODO: Implement YAML rendering
    #    raise NotImplementedError

    # def render_as_yaml(self, name):
    #    '''
    #    Render the configuration item as a parseable YAML string including comments
    #    '''
    #    # TODO: Include the item rules in the output, minimum, maximum, etc...
    #    output = '# ----- '
    #    output += self.title
    #    output += ' '
    #    output += '-' * (RENDER_COMMENT_YAML_MAX_LINE_LENGTH - 7 - len(self.title) - 2)
    #    output += '>\n'
    #    if self.description:
    #        output += '\n'.join(textwrap.wrap(self.description,
    #                                          width=RENDER_COMMENT_YAML_MAX_LINE_LENGTH,
    #                                          initial_indent='# '))
    #        output += '\n'
    #        yamled_default_value = salt.utils.yaml.safe_dump(self.default, default_flow_style=False).split('\n...', 1)[0]
    #        output += '# Default: {0}\n'.format(yamled_default_value)
    #        output += '#{0}: {1}\n'.format(name, yamled_default_value)
    #    output += '# <---- '
    #    output += self.title
    #    output += ' '
    #    output += '-' * (RENDER_COMMENT_YAML_MAX_LINE_LENGTH - 7 - len(self.title) - 1)
    #    return output + '\n'


class NullItem(BaseSchemaItem):

    __type__ = "null"


class BooleanItem(BaseSchemaItem):
    __type__ = "boolean"


class StringItem(BaseSchemaItem):
    """
    A string configuration field
    """

    __type__ = "string"

    __serialize_attr_aliases__ = {"min_length": "minLength", "max_length": "maxLength"}

    format = None
    pattern = None
    min_length = None
    max_length = None

    def __init__(
        self,
        format=None,  # pylint: disable=redefined-builtin
        pattern=None,
        min_length=None,
        max_length=None,
        **kwargs
    ):
        """
        :param required:
            If the configuration item is required. Defaults to ``False``.
        :param title:
            A short explanation about the purpose of the data described by this item.
        :param description:
            A detailed explanation about the purpose of the data described by this item.
        :param default:
            The default value for this configuration item. May be :data:`.Null` (a special value
            to set the default value to null).
        :param enum:
            A list(list, tuple, set) of valid choices.
        :param format:
            A semantic format of the string (for example, ``"date-time"``, ``"email"``, or ``"uri"``).
        :param pattern:
            A regular expression (ECMA 262) that a string value must match.
        :param min_length:
            The minimum length
        :param max_length:
            The maximum length
        """
        if format is not None:  # pylint: disable=redefined-builtin
            self.format = format
        if pattern is not None:
            self.pattern = pattern
        if min_length is not None:
            self.min_length = min_length
        if max_length is not None:
            self.max_length = max_length
        super().__init__(**kwargs)

    def __validate_attributes__(self):
        if self.format is None and self.__format__ is not None:
            self.format = self.__format__


class EMailItem(StringItem):
    """
    An internet email address, see `RFC 5322, section 3.4.1`__.

    .. __: http://tools.ietf.org/html/rfc5322
    """

    __format__ = "email"


class IPv4Item(StringItem):
    """
    An IPv4 address configuration field, according to dotted-quad ABNF syntax as defined in
    `RFC 2673, section 3.2`__.

    .. __: http://tools.ietf.org/html/rfc2673
    """

    __format__ = "ipv4"


class IPv6Item(StringItem):
    """
    An IPv6 address configuration field, as defined in `RFC 2373, section 2.2`__.

    .. __: http://tools.ietf.org/html/rfc2373
    """

    __format__ = "ipv6"


class HostnameItem(StringItem):
    """
    An Internet host name configuration field, see `RFC 1034, section 3.1`__.

    .. __: http://tools.ietf.org/html/rfc1034
    """

    __format__ = "hostname"


class DateTimeItem(StringItem):
    """
    An ISO 8601 formatted date-time configuration field, as defined by `RFC 3339, section 5.6`__.

    .. __: http://tools.ietf.org/html/rfc3339
    """

    __format__ = "date-time"


class UriItem(StringItem):
    """
    A universal resource identifier (URI) configuration field, according to `RFC3986`__.

    .. __: http://tools.ietf.org/html/rfc3986
    """

    __format__ = "uri"


class SecretItem(StringItem):
    """
    A string configuration field containing a secret, for example, passwords, API keys, etc
    """

    __format__ = "secret"


class NumberItem(BaseSchemaItem):

    __type__ = "number"

    __serialize_attr_aliases__ = {
        "multiple_of": "multipleOf",
        "exclusive_minimum": "exclusiveMinimum",
        "exclusive_maximum": "exclusiveMaximum",
    }

    multiple_of = None
    minimum = None
    exclusive_minimum = None
    maximum = None
    exclusive_maximum = None

    def __init__(
        self,
        multiple_of=None,
        minimum=None,
        exclusive_minimum=None,
        maximum=None,
        exclusive_maximum=None,
        **kwargs
    ):
        """
        :param required:
            If the configuration item is required. Defaults to ``False``.
        :param title:
            A short explanation about the purpose of the data described by this item.
        :param description:
            A detailed explanation about the purpose of the data described by this item.
        :param default:
            The default value for this configuration item. May be :data:`.Null` (a special value
            to set the default value to null).
        :param enum:
            A list(list, tuple, set) of valid choices.
        :param multiple_of:
            A value must be a multiple of this factor.
        :param minimum:
            The minimum allowed value
        :param exclusive_minimum:
            Whether a value is allowed to be exactly equal to the minimum
        :param maximum:
            The maximum allowed value
        :param exclusive_maximum:
            Whether a value is allowed to be exactly equal to the maximum
        """
        if multiple_of is not None:
            self.multiple_of = multiple_of
        if minimum is not None:
            self.minimum = minimum
        if exclusive_minimum is not None:
            self.exclusive_minimum = exclusive_minimum
        if maximum is not None:
            self.maximum = maximum
        if exclusive_maximum is not None:
            self.exclusive_maximum = exclusive_maximum
        super().__init__(**kwargs)


class IntegerItem(NumberItem):
    __type__ = "integer"


class ArrayItem(BaseSchemaItem):
    __type__ = "array"

    __serialize_attr_aliases__ = {
        "min_items": "minItems",
        "max_items": "maxItems",
        "unique_items": "uniqueItems",
        "additional_items": "additionalItems",
    }

    items = None
    min_items = None
    max_items = None
    unique_items = None
    additional_items = None

    def __init__(
        self,
        items=None,
        min_items=None,
        max_items=None,
        unique_items=None,
        additional_items=None,
        **kwargs
    ):
        """
        :param required:
            If the configuration item is required. Defaults to ``False``.
        :param title:
            A short explanation about the purpose of the data described by this item.
        :param description:
            A detailed explanation about the purpose of the data described by this item.
        :param default:
            The default value for this configuration item. May be :data:`.Null` (a special value
            to set the default value to null).
        :param enum:
            A list(list, tuple, set) of valid choices.
        :param items:
            Either of the following:
                * :class:`BaseSchemaItem` -- all items of the array must match the field schema;
                * a list or a tuple of :class:`fields <.BaseSchemaItem>` -- all items of the array must be
                  valid according to the field schema at the corresponding index (tuple typing);
        :param min_items:
            Minimum length of the array
        :param max_items:
            Maximum length of the array
        :param unique_items:
            Whether all the values in the array must be distinct.
        :param additional_items:
            If the value of ``items`` is a list or a tuple, and the array length is larger than
            the number of fields in ``items``, then the additional items are described
            by the :class:`.BaseField` passed using this argument.
        :type additional_items: bool or :class:`.BaseSchemaItem`
        """
        if items is not None:
            self.items = items
        if min_items is not None:
            self.min_items = min_items
        if max_items is not None:
            self.max_items = max_items
        if unique_items is not None:
            self.unique_items = unique_items
        if additional_items is not None:
            self.additional_items = additional_items
        super().__init__(**kwargs)

    def __validate_attributes__(self):
        if not self.items and not self.additional_items:
            raise RuntimeError("One of items or additional_items must be passed.")
        if self.items is not None:
            if isinstance(self.items, (list, tuple)):
                for item in self.items:
                    if not isinstance(item, (Schema, SchemaItem)):
                        raise RuntimeError(
                            "All items passed in the item argument tuple/list must be "
                            "a subclass of Schema, SchemaItem or BaseSchemaItem, "
                            "not {}".format(type(item))
                        )
            elif not isinstance(self.items, (Schema, SchemaItem)):
                raise RuntimeError(
                    "The items argument passed must be a subclass of "
                    "Schema, SchemaItem or BaseSchemaItem, not "
                    "{}".format(type(self.items))
                )

    def __get_items__(self):
        if isinstance(self.items, (Schema, SchemaItem)):
            # This is either a Schema or a Basetem, return it in its
            # serialized form
            return self.items.serialize()
        if isinstance(self.items, (tuple, list)):
            items = []
            for item in self.items:
                items.append(item.serialize())
            return items


class DictItem(BaseSchemaItem):

    __type__ = "object"

    __serialize_attr_aliases__ = {
        "min_properties": "minProperties",
        "max_properties": "maxProperties",
        "pattern_properties": "patternProperties",
        "additional_properties": "additionalProperties",
    }

    properties = None
    pattern_properties = None
    additional_properties = None
    min_properties = None
    max_properties = None

    def __init__(
        self,
        properties=None,
        pattern_properties=None,
        additional_properties=None,
        min_properties=None,
        max_properties=None,
        **kwargs
    ):
        """
        :param required:
            If the configuration item is required. Defaults to ``False``.
        :type required:
            boolean
        :param title:
            A short explanation about the purpose of the data described by this item.
        :type title:
            str
        :param description:
            A detailed explanation about the purpose of the data described by this item.
        :param default:
            The default value for this configuration item. May be :data:`.Null` (a special value
            to set the default value to null).
        :param enum:
            A list(list, tuple, set) of valid choices.
        :param properties:
            A dictionary containing fields
        :param pattern_properties:
            A dictionary whose keys are regular expressions (ECMA 262).
            Properties match against these regular expressions, and for any that match,
            the property is described by the corresponding field schema.
        :type pattern_properties: dict[str -> :class:`.Schema` or
                                       :class:`.SchemaItem` or :class:`.BaseSchemaItem`]
        :param additional_properties:
            Describes properties that are not described by the ``properties`` or ``pattern_properties``.
        :type additional_properties: bool or :class:`.Schema` or :class:`.SchemaItem`
                                     or :class:`.BaseSchemaItem`
        :param min_properties:
            A minimum number of properties.
        :type min_properties: int
        :param max_properties:
            A maximum number of properties
        :type max_properties: int
        """
        if properties is not None:
            self.properties = properties
        if pattern_properties is not None:
            self.pattern_properties = pattern_properties
        if additional_properties is not None:
            self.additional_properties = additional_properties
        if min_properties is not None:
            self.min_properties = min_properties
        if max_properties is not None:
            self.max_properties = max_properties
        super().__init__(**kwargs)

    def __validate_attributes__(self):
        if (
            not self.properties
            and not self.pattern_properties
            and not self.additional_properties
        ):
            raise RuntimeError(
                "One of properties, pattern_properties or additional_properties must be"
                " passed"
            )
        if self.properties is not None:
            if not isinstance(self.properties, (Schema, dict)):
                raise RuntimeError(
                    "The passed properties must be passed as a dict or "
                    " a Schema not '{}'".format(type(self.properties))
                )
            if not isinstance(self.properties, Schema):
                for key, prop in self.properties.items():
                    if not isinstance(prop, (Schema, SchemaItem)):
                        raise RuntimeError(
                            "The passed property who's key is '{}' must be of type "
                            "Schema, SchemaItem or BaseSchemaItem, not "
                            "'{}'".format(key, type(prop))
                        )
        if self.pattern_properties is not None:
            if not isinstance(self.pattern_properties, dict):
                raise RuntimeError(
                    "The passed pattern_properties must be passed as a dict "
                    "not '{}'".format(type(self.pattern_properties))
                )
            for key, prop in self.pattern_properties.items():
                if not isinstance(prop, (Schema, SchemaItem)):
                    raise RuntimeError(
                        "The passed pattern_property who's key is '{}' must "
                        "be of type Schema, SchemaItem or BaseSchemaItem, "
                        "not '{}'".format(key, type(prop))
                    )
        if self.additional_properties is not None:
            if not isinstance(self.additional_properties, (bool, Schema, SchemaItem)):
                raise RuntimeError(
                    "The passed additional_properties must be of type bool, "
                    "Schema, SchemaItem or BaseSchemaItem, not '{}'".format(
                        type(self.pattern_properties)
                    )
                )

    def __get_properties__(self):
        if self.properties is None:
            return
        if isinstance(self.properties, Schema):
            return self.properties.serialize()["properties"]
        properties = OrderedDict()
        for key, prop in self.properties.items():
            properties[key] = prop.serialize()
        return properties

    def __get_pattern_properties__(self):
        if self.pattern_properties is None:
            return
        pattern_properties = OrderedDict()
        for key, prop in self.pattern_properties.items():
            pattern_properties[key] = prop.serialize()
        return pattern_properties

    def __get_additional_properties__(self):
        if self.additional_properties is None:
            return
        if isinstance(self.additional_properties, bool):
            return self.additional_properties
        return self.additional_properties.serialize()

    def __call__(self, flatten=False):
        self.__flatten__ = flatten
        return self

    def serialize(self):
        result = super().serialize()
        required = []
        if self.properties is not None:
            if isinstance(self.properties, Schema):
                serialized = self.properties.serialize()
                if "required" in serialized:
                    required.extend(serialized["required"])
            else:
                for key, prop in self.properties.items():
                    if prop.required:
                        required.append(key)
        if required:
            result["required"] = required
        return result


class RequirementsItem(SchemaItem):
    __type__ = "object"

    requirements = None

    def __init__(self, requirements=None):
        if requirements is not None:
            self.requirements = requirements
        super().__init__()

    def __validate_attributes__(self):
        if self.requirements is None:
            raise RuntimeError("The passed requirements must not be empty")
        if not isinstance(self.requirements, (SchemaItem, list, tuple, set)):
            raise RuntimeError(
                "The passed requirements must be passed as a list, tuple, "
                "set SchemaItem or BaseSchemaItem, not '{}'".format(self.requirements)
            )

        if not isinstance(self.requirements, SchemaItem):
            if not isinstance(self.requirements, list):
                self.requirements = list(self.requirements)

            for idx, item in enumerate(self.requirements):
                if not isinstance(item, ((str,), SchemaItem)):
                    raise RuntimeError(
                        "The passed requirement at the {} index must be of type "
                        "str or SchemaItem, not '{}'".format(idx, type(item))
                    )

    def serialize(self):
        if isinstance(self.requirements, SchemaItem):
            requirements = self.requirements.serialize()
        else:
            requirements = []
            for requirement in self.requirements:
                if isinstance(requirement, SchemaItem):
                    requirements.append(requirement.serialize())
                    continue
                requirements.append(requirement)
        return {"required": requirements}


class OneOfItem(SchemaItem):

    __type__ = "oneOf"

    items = None

    def __init__(self, items=None, required=None):
        if items is not None:
            self.items = items
        super().__init__(required=required)

    def __validate_attributes__(self):
        if not self.items:
            raise RuntimeError("The passed items must not be empty")
        if not isinstance(self.items, (list, tuple)):
            raise RuntimeError(
                "The passed items must be passed as a list/tuple not '{}'".format(
                    type(self.items)
                )
            )
        for idx, item in enumerate(self.items):
            if not isinstance(item, (Schema, SchemaItem)):
                raise RuntimeError(
                    "The passed item at the {} index must be of type "
                    "Schema, SchemaItem or BaseSchemaItem, not "
                    "'{}'".format(idx, type(item))
                )
        if not isinstance(self.items, list):
            self.items = list(self.items)

    def __call__(self, flatten=False):
        self.__flatten__ = flatten
        return self

    def serialize(self):
        return {self.__type__: [i.serialize() for i in self.items]}


class AnyOfItem(OneOfItem):

    __type__ = "anyOf"


class AllOfItem(OneOfItem):

    __type__ = "allOf"


class NotItem(SchemaItem):

    __type__ = "not"

    item = None

    def __init__(self, item=None):
        if item is not None:
            self.item = item
        super().__init__()

    def __validate_attributes__(self):
        if not self.item:
            raise RuntimeError("An item must be passed")
        if not isinstance(self.item, (Schema, SchemaItem)):
            raise RuntimeError(
                "The passed item be of type Schema, SchemaItem or "
                "BaseSchemaItem, not '{}'".format(type(self.item))
            )

    def serialize(self):
        return {self.__type__: self.item.serialize()}


# ----- Custom Preconfigured Configs -------------------------------------------------------------------------------->
class PortItem(IntegerItem):
    minimum = 0  # yes, 0 is a valid port number
    maximum = 65535


# <---- Custom Preconfigured Configs ---------------------------------------------------------------------------------


class ComplexSchemaItem(BaseSchemaItem):
    """
    .. versionadded:: 2016.11.0

    Complex Schema Item
    """

    # This attribute is populated by the metaclass, but pylint fails to see it
    # and assumes it's not an iterable
    _attributes = []
    _definition_name = None

    def __init__(self, definition_name=None, required=None):
        super().__init__(required=required)
        self.__type__ = "object"
        self._definition_name = (
            definition_name if definition_name else self.__class__.__name__
        )
        # Schema attributes might have been added as class attributes so we
        # and they must be added to the _attributes attr
        self._add_missing_schema_attributes()

    def _add_missing_schema_attributes(self):
        """
        Adds any missed schema attributes to the _attributes list

        The attributes can be class attributes and they won't be
        included in the _attributes list automatically
        """
        for attr in [attr for attr in dir(self) if not attr.startswith("__")]:
            attr_val = getattr(self, attr)
            if (
                isinstance(getattr(self, attr), SchemaItem)
                and attr not in self._attributes
            ):

                self._attributes.append(attr)

    @property
    def definition_name(self):
        return self._definition_name

    def serialize(self):
        """
        The serialization of the complex item is a pointer to the item
        definition
        """
        return {"$ref": "#/definitions/{}".format(self.definition_name)}

    def get_definition(self):
        """Returns the definition of the complex item"""

        serialized = super().serialize()
        # Adjust entries in the serialization
        del serialized["definition_name"]
        serialized["title"] = self.definition_name

        properties = {}
        required_attr_names = []

        for attr_name in self._attributes:
            attr = getattr(self, attr_name)
            if attr and isinstance(attr, BaseSchemaItem):
                # Remove the attribute entry added by the base serialization
                del serialized[attr_name]
                properties[attr_name] = attr.serialize()
                properties[attr_name]["type"] = attr.__type__
                if attr.required:
                    required_attr_names.append(attr_name)
        if serialized.get("properties") is None:
            serialized["properties"] = {}
        serialized["properties"].update(properties)

        # Assign the required array
        if required_attr_names:
            serialized["required"] = required_attr_names
        return serialized

    def get_complex_attrs(self):
        """Returns a dictionary of the complex attributes"""
        return [
            getattr(self, attr_name)
            for attr_name in self._attributes
            if isinstance(getattr(self, attr_name), ComplexSchemaItem)
        ]


class DefinitionsSchema(Schema):
    """
    .. versionadded:: 2016.11.0

    JSON schema class that supports ComplexSchemaItem objects by adding
    a definitions section to the JSON schema, containing the item definitions.

    All references to ComplexSchemaItems are built using schema inline
    dereferencing.
    """

    @classmethod
    def serialize(cls, id_=None):
        # Get the initial serialization
        serialized = super().serialize(id_)
        complex_items = []
        # Augment the serializations with the definitions of all complex items
        aux_items = cls._items.values()

        # Convert dict_view object to a list on Python 3
        aux_items = list(aux_items)

        while aux_items:
            item = aux_items.pop(0)
            # Add complex attributes
            if isinstance(item, ComplexSchemaItem):
                complex_items.append(item)
                aux_items.extend(item.get_complex_attrs())

            # Handle container items
            if isinstance(item, OneOfItem):
                aux_items.extend(item.items)
            elif isinstance(item, ArrayItem):
                aux_items.append(item.items)
            elif isinstance(item, DictItem):
                if item.properties:
                    aux_items.extend(item.properties.values())
                if item.additional_properties and isinstance(
                    item.additional_properties, SchemaItem
                ):

                    aux_items.append(item.additional_properties)

        definitions = OrderedDict()
        for config in complex_items:
            if isinstance(config, ComplexSchemaItem):
                definitions[config.definition_name] = config.get_definition()
        serialized["definitions"] = definitions
        return serialized



"""
Send events covering process status
"""
import logging

import salt.utils.beacons

try:
    import salt.utils.psutil_compat as psutil

    HAS_PSUTIL = True
except ImportError:
    HAS_PSUTIL = False


log = logging.getLogger(__name__)  # pylint: disable=invalid-name

__virtualname__ = "ps"


def __virtual__():
    if not HAS_PSUTIL:
        err_msg = "psutil library is missing."
        log.error("Unable to load %s beacon: %s", __virtualname__, err_msg)
        return False, err_msg
    return __virtualname__


def validate(config):
    """
    Validate the beacon configuration
    """
    # Configuration for ps beacon should be a list of dicts
    if not isinstance(config, list):
        return False, "Configuration for ps beacon must be a list."
    else:
        config = salt.utils.beacons.list_to_dict(config)

        if "processes" not in config:
            return False, "Configuration for ps beacon requires processes."
        else:
            if not isinstance(config["processes"], dict):
                return False, "Processes for ps beacon must be a dictionary."

    return True, "Valid beacon configuration"


def beacon(config):
    """
    Scan for processes and fire events

    Example Config

    .. code-block:: yaml

        beacons:
          ps:
            - processes:
                salt-master: running
                mysql: stopped

    The config above sets up beacons to check that
    processes are running or stopped.
    """
    ret = []
    procs = []
    for proc in psutil.process_iter():
        try:
            _name = proc.name()
        except psutil.NoSuchProcess:
            # The process is now gone
            continue
        if _name not in procs:
            procs.append(_name)

    config = salt.utils.beacons.list_to_dict(config)

    for process in config.get("processes", {}):
        ret_dict = {}
        if config["processes"][process] == "running":
            if process in procs:
                ret_dict[process] = "Running"
                ret.append(ret_dict)
        elif config["processes"][process] == "stopped":
            if process not in procs:
                ret_dict[process] = "Stopped"
                ret.append(ret_dict)
        else:
            if process not in procs:
                ret_dict[process] = False
                ret.append(ret_dict)
    return ret



"""
    salt.utils.yamldumper
    ~~~~~~~~~~~~~~~~~~~~~

"""
# pylint: disable=W0232
#         class has no __init__ method


import collections

import yaml  # pylint: disable=blacklisted-import

import salt.utils.context
from salt.utils.odict import OrderedDict

try:
    from yaml import CDumper as Dumper
    from yaml import CSafeDumper as SafeDumper
except ImportError:
    from yaml import Dumper, SafeDumper


__all__ = [
    "OrderedDumper",
    "SafeOrderedDumper",
    "IndentedSafeOrderedDumper",
    "get_dumper",
    "dump",
    "safe_dump",
]


class IndentMixin(Dumper):
    """
    Mixin that improves YAML dumped list readability
    by indenting them by two spaces,
    instead of being flush with the key they are under.
    """

    def increase_indent(self, flow=False, indentless=False):
        return super().increase_indent(flow, False)


class OrderedDumper(Dumper):
    """
    A YAML dumper that represents python OrderedDict as simple YAML map.
    """


class SafeOrderedDumper(SafeDumper):
    """
    A YAML safe dumper that represents python OrderedDict as simple YAML map.
    """


class IndentedSafeOrderedDumper(IndentMixin, SafeOrderedDumper):
    """
    A YAML safe dumper that represents python OrderedDict as simple YAML map,
    and also indents lists by two spaces.
    """


def represent_ordereddict(dumper, data):
    return dumper.represent_dict(list(data.items()))


def represent_undefined(dumper, data):
    return dumper.represent_scalar("tag:yaml.org,2002:null", "NULL")


OrderedDumper.add_representer(OrderedDict, represent_ordereddict)
SafeOrderedDumper.add_representer(OrderedDict, represent_ordereddict)
SafeOrderedDumper.add_representer(None, represent_undefined)

OrderedDumper.add_representer(
    collections.defaultdict, yaml.representer.SafeRepresenter.represent_dict
)
SafeOrderedDumper.add_representer(
    collections.defaultdict, yaml.representer.SafeRepresenter.represent_dict
)
OrderedDumper.add_representer(
    salt.utils.context.NamespacedDictWrapper,
    yaml.representer.SafeRepresenter.represent_dict,
)
SafeOrderedDumper.add_representer(
    salt.utils.context.NamespacedDictWrapper,
    yaml.representer.SafeRepresenter.represent_dict,
)

OrderedDumper.add_representer(
    "tag:yaml.org,2002:timestamp", OrderedDumper.represent_scalar
)
SafeOrderedDumper.add_representer(
    "tag:yaml.org,2002:timestamp", SafeOrderedDumper.represent_scalar
)


def get_dumper(dumper_name):
    return {
        "OrderedDumper": OrderedDumper,
        "SafeOrderedDumper": SafeOrderedDumper,
        "IndentedSafeOrderedDumper": IndentedSafeOrderedDumper,
    }.get(dumper_name)


def dump(data, stream=None, **kwargs):
    """
    .. versionadded:: 2018.3.0

    Helper that wraps yaml.dump and ensures that we encode unicode strings
    unless explicitly told not to.
    """
    if "allow_unicode" not in kwargs:
        kwargs["allow_unicode"] = True
    kwargs.setdefault("default_flow_style", None)
    return yaml.dump(data, stream, **kwargs)


def safe_dump(data, stream=None, **kwargs):
    """
    Use a custom dumper to ensure that defaultdict and OrderedDict are
    represented properly. Ensure that unicode strings are encoded unless
    explicitly told not to.
    """
    if "allow_unicode" not in kwargs:
        kwargs["allow_unicode"] = True
    kwargs.setdefault("default_flow_style", None)
    return yaml.dump(data, stream, Dumper=SafeOrderedDumper, **kwargs)



from contextlib import ExitStack

import pytest

pytestmark = [
    pytest.mark.slow_test,
]


def get_module_types():
    module_types = [
        "clouds",
        "modules",
        "states",
        "grains",
        "renderers",
        "returners",
        "output",
        "proxy",
        "runners",
        "wheel",
        "engines",
        "thorium",
        "queues",
        "pillar",
        "utils",
        "sdb",
        "cache",
        "fileserver",
        "tops",
        "tokens",
        "serializers",
        "executors",
        "roster",
    ]
    return module_types


@pytest.fixture(params=get_module_types())
def module_type(request):
    yield request.param


@pytest.fixture
def module_sync_functions():
    yield {
        "clouds": "clouds",
        "modules": "modules",
        "states": "states",
        "grains": "grains",
        "renderers": "renderers",
        "returners": "returners",
        "output": "output",
        "proxy": "proxymodules",
        "runners": "runners",
        "wheel": "wheel",
        "engines": "engines",
        "thorium": "thorium",
        "queues": "queues",
        "pillar": "pillar",
        "utils": "utils",
        "sdb": "sdb",
        "cache": "cache",
        "fileserver": "fileserver",
        "tops": "tops",
        "tokens": "eauth_tokens",
        "serializers": "serializers",
        "executors": "executors",
        "roster": "roster",
    }


def test_sync(
    module_type, module_sync_functions, salt_run_cli, salt_minion, salt_master
):
    """
    Ensure modules are synced when various sync functions are called
    """
    module_name = "hello_sync_{}".format(module_type)
    module_contents = """
def __virtual__():
    return "hello"

def world():
    return "world"
"""

    test_moduledir = salt_master.state_tree.base.write_path / "_{}".format(module_type)
    test_moduledir.mkdir(parents=True, exist_ok=True)
    module_tempfile = salt_master.state_tree.base.temp_file(
        "_{}/{}.py".format(module_type, module_name), module_contents
    )

    with module_tempfile:
        salt_cmd = "saltutil.sync_{}".format(module_sync_functions[module_type])
        ret = salt_run_cli.run(salt_cmd)
        assert ret.returncode == 0
        assert "{}.hello".format(module_type) in ret.stdout


def _write_module_dir_and_file(module_type, salt_minion, salt_master):
    """
    Write out dummy module to appropriate module location
    """
    module_name = "hello_sync_all"
    module_contents = """
def __virtual__():
    return "hello"

def world():
    return "world"
"""

    test_moduledir = salt_master.state_tree.base.paths[0] / "_{}".format(module_type)
    test_moduledir.mkdir(parents=True, exist_ok=True)

    module_tempfile = salt_master.state_tree.base.temp_file(
        "_{}/{}.py".format(module_type, module_name), module_contents
    )

    return module_tempfile


def test_sync_all(salt_run_cli, salt_minion, salt_master):
    """
    Ensure all modules are synced when sync_all function is called
    """

    with ExitStack() as stack:
        files = [
            stack.enter_context(
                _write_module_dir_and_file(_module_type, salt_minion, salt_master)
            )
            for _module_type in get_module_types()
        ]

        salt_cmd = "saltutil.sync_all"
        ret = salt_run_cli.run(salt_cmd)

        assert ret.returncode == 0
        for module_type in get_module_types():
            assert "{}.hello".format(module_type) in ret.stdout



import os

import pytest


@pytest.fixture(scope="module")
def install():
    pytest.helpers.clean_env()

    # Create an existing config
    pytest.helpers.existing_config()

    # Create a custom config
    pytest.helpers.custom_config()

    pytest.helpers.run_command(
        [
            pytest.INST_BIN,
            "/custom-config=custom_conf",
            "/master=cli_master",
            "/minion-name=cli_minion",
        ]
    )
    yield
    pytest.helpers.clean_env()


def test_binaries_present(install):
    assert os.path.exists(f"{pytest.INST_DIR}\\ssm.exe")


def test_config_present(install):
    assert os.path.exists(f"{pytest.DATA_DIR}\\conf\\minion")


def test_config_correct(install):
    # The config file should be the custom config with master and minion set
    expected = [
        "# Custom config from test suite line 1/6\n",
        "master: cli_master\n",
        "# Custom config from test suite line 2/6\n",
        "id: cli_minion\n",
        "# Custom config from test suite line 3/6\n",
        "# Custom config from test suite line 4/6\n",
        "# Custom config from test suite line 5/6\n",
        "# Custom config from test suite line 6/6\n",
    ]

    with open(f"{pytest.DATA_DIR}\\conf\\minion") as f:
        result = f.readlines()

    assert result == expected



"""
Various user validation utilities
"""

import logging
import re

log = logging.getLogger(__name__)

VALID_USERNAME = re.compile(r"[a-z_][a-z0-9_-]*[$]?", re.IGNORECASE)


def valid_username(user):
    """
    Validates a username based on the guidelines in `useradd(8)`
    """
    if not isinstance(user, str):
        return False

    if len(user) > 32:
        return False

    return VALID_USERNAME.match(user) is not None



"""
    :codeauthor: :email:`Anthony Shaw <anthonyshaw@apache.org>`
"""

import pytest

import salt.modules.napalm_ntp as napalm_ntp
import tests.support.napalm as napalm_test_support
from tests.support.mock import MagicMock, patch


def mock_net_load_template(template, *args, **kwargs):
    if template == "set_ntp_peers" or template == "delete_ntp_peers":
        assert "1.2.3.4" in kwargs["peers"]
    if template == "set_ntp_servers" or template == "delete_ntp_servers":
        assert "2.2.3.4" in kwargs["servers"]


@pytest.fixture
def configure_loader_modules():
    module_globals = {
        "__salt__": {
            "config.get": MagicMock(
                return_value={"test": {"driver": "test", "key": "2orgk34kgk34g"}}
            ),
            "file.file_exists": napalm_test_support.true,
            "file.join": napalm_test_support.join,
            "file.get_managed": napalm_test_support.get_managed_file,
            "random.hash": napalm_test_support.random_hash,
            "net.load_template": mock_net_load_template,
        }
    }

    return {napalm_ntp: module_globals}


def test_peers():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_ntp.peers()
        assert "172.17.17.1" in ret["out"]


def test_servers():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_ntp.servers()
        assert "172.17.17.1" in ret["out"]


def test_stats():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_ntp.stats()
        assert ret["out"][0]["reachability"] == 377


def test_set_peers():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_ntp.set_peers("1.2.3.4", "5.6.7.8")
        assert ret is None


def test_set_servers():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_ntp.set_servers("2.2.3.4", "6.6.7.8")
        assert ret is None


def test_delete_servers():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_ntp.delete_servers("2.2.3.4", "6.6.7.8")
        assert ret is None


def test_delete_peers():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_ntp.delete_peers("1.2.3.4", "5.6.7.8")
        assert ret is None



"""
    salt.serializers.json
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implements JSON serializer.

    It's just a wrapper around json (or simplejson if available).
"""


import salt.utils.json
from salt.serializers import DeserializationError, SerializationError

try:
    import simplejson as _json
except ImportError:
    import json as _json  # pylint: disable=blacklisted-import


__all__ = ["deserialize", "serialize", "available"]

available = True


def deserialize(stream_or_string, **options):
    """
    Deserialize any string or stream like object into a Python data structure.

    :param stream_or_string: stream or string to deserialize.
    :param options: options given to lower json/simplejson module.
    """

    try:
        if not isinstance(stream_or_string, (bytes, str)):
            return salt.utils.json.load(stream_or_string, _json_module=_json, **options)

        if isinstance(stream_or_string, bytes):
            stream_or_string = stream_or_string.decode("utf-8")

        return salt.utils.json.loads(stream_or_string, _json_module=_json)
    except Exception as error:  # pylint: disable=broad-except
        raise DeserializationError(error)


def serialize(obj, **options):
    """
    Serialize Python data to JSON.

    :param obj: the data structure to serialize
    :param options: options given to lower json/simplejson module.
    """

    try:
        if "fp" in options:
            return salt.utils.json.dump(obj, _json_module=_json, **options)
        else:
            return salt.utils.json.dumps(obj, _json_module=_json, **options)
    except Exception as error:  # pylint: disable=broad-except
        raise SerializationError(error)



"""
Connection module for Amazon CloudTrail

.. versionadded:: 2016.3.0

:depends:
    - boto
    - boto3

The dependencies listed above can be installed via package or pip.

:configuration: This module accepts explicit Lambda credentials but can also
    utilize IAM roles assigned to the instance through Instance Profiles.
    Dynamic credentials are then automatically obtained from AWS API and no
    further configuration is necessary. More Information available at:

    .. code-block:: text

        http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html

    If IAM roles are not used you need to specify them either in a pillar or
    in the minion's config file:

    .. code-block:: yaml

        cloudtrail.keyid: GKTADJGHEIQSXMKKRBJ08H
        cloudtrail.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

    A region may also be specified in the configuration:

    .. code-block:: yaml

        cloudtrail.region: us-east-1

    If a region is not specified, the default is us-east-1.

    It's also possible to specify key, keyid and region via a profile, either
    as a passed in dict, or as a string to pull from pillars or minion config:

    .. code-block:: yaml

        myprofile:
            keyid: GKTADJGHEIQSXMKKRBJ08H
            key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
            region: us-east-1

"""
# keep lint from choking on _get_conn and _cache_id
# pylint: disable=E0602


import logging

import salt.utils.compat
import salt.utils.versions

log = logging.getLogger(__name__)


# pylint: disable=import-error
try:
    # pylint: disable=unused-import
    import boto
    import boto3

    # pylint: enable=unused-import
    from botocore.exceptions import ClientError

    logging.getLogger("boto3").setLevel(logging.CRITICAL)
    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False
# pylint: enable=import-error


def __virtual__():
    """
    Only load if boto libraries exist and if boto libraries are greater than
    a given version.
    """
    # the boto_lambda execution module relies on the connect_to_region() method
    # which was added in boto 2.8.0
    # https://github.com/boto/boto/commit/33ac26b416fbb48a60602542b4ce15dcc7029f12
    return salt.utils.versions.check_boto_reqs(boto3_ver="1.2.5")


def __init__(opts):
    if HAS_BOTO:
        __utils__["boto3.assign_funcs"](__name__, "cloudtrail")


def exists(Name, region=None, key=None, keyid=None, profile=None):
    """
    Given a trail name, check to see if the given trail exists.

    Returns True if the given trail exists and returns False if the given
    trail does not exist.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_cloudtrail.exists mytrail

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        conn.get_trail_status(Name=Name)
        return {"exists": True}
    except ClientError as e:
        err = __utils__["boto3.get_error"](e)
        if e.response.get("Error", {}).get("Code") == "TrailNotFoundException":
            return {"exists": False}
        return {"error": err}


def create(
    Name,
    S3BucketName,
    S3KeyPrefix=None,
    SnsTopicName=None,
    IncludeGlobalServiceEvents=None,
    IsMultiRegionTrail=None,
    EnableLogFileValidation=None,
    CloudWatchLogsLogGroupArn=None,
    CloudWatchLogsRoleArn=None,
    KmsKeyId=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Given a valid config, create a trail.

    Returns {created: true} if the trail was created and returns
    {created: False} if the trail was not created.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_cloudtrail.create my_trail my_bucket

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        kwargs = {}
        for arg in (
            "S3KeyPrefix",
            "SnsTopicName",
            "IncludeGlobalServiceEvents",
            "IsMultiRegionTrail",
            "EnableLogFileValidation",
            "CloudWatchLogsLogGroupArn",
            "CloudWatchLogsRoleArn",
            "KmsKeyId",
        ):
            if locals()[arg] is not None:
                kwargs[arg] = locals()[arg]
        trail = conn.create_trail(Name=Name, S3BucketName=S3BucketName, **kwargs)
        if trail:
            log.info("The newly created trail name is %s", trail["Name"])

            return {"created": True, "name": trail["Name"]}
        else:
            log.warning("Trail was not created")
            return {"created": False}
    except ClientError as e:
        return {"created": False, "error": __utils__["boto3.get_error"](e)}


def delete(Name, region=None, key=None, keyid=None, profile=None):
    """
    Given a trail name, delete it.

    Returns {deleted: true} if the trail was deleted and returns
    {deleted: false} if the trail was not deleted.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_cloudtrail.delete mytrail

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        conn.delete_trail(Name=Name)
        return {"deleted": True}
    except ClientError as e:
        return {"deleted": False, "error": __utils__["boto3.get_error"](e)}


def describe(Name, region=None, key=None, keyid=None, profile=None):
    """
    Given a trail name describe its properties.

    Returns a dictionary of interesting properties.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_cloudtrail.describe mytrail

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        trails = conn.describe_trails(trailNameList=[Name])
        if trails and trails.get("trailList"):
            keys = (
                "Name",
                "S3BucketName",
                "S3KeyPrefix",
                "SnsTopicName",
                "IncludeGlobalServiceEvents",
                "IsMultiRegionTrail",
                "HomeRegion",
                "TrailARN",
                "LogFileValidationEnabled",
                "CloudWatchLogsLogGroupArn",
                "CloudWatchLogsRoleArn",
                "KmsKeyId",
            )
            trail = trails["trailList"].pop()
            return {"trail": {k: trail.get(k) for k in keys}}
        else:
            return {"trail": None}
    except ClientError as e:
        err = __utils__["boto3.get_error"](e)
        if e.response.get("Error", {}).get("Code") == "TrailNotFoundException":
            return {"trail": None}
        return {"error": __utils__["boto3.get_error"](e)}


def status(Name, region=None, key=None, keyid=None, profile=None):
    """
    Given a trail name describe its properties.

    Returns a dictionary of interesting properties.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_cloudtrail.describe mytrail

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        trail = conn.get_trail_status(Name=Name)
        if trail:
            keys = (
                "IsLogging",
                "LatestDeliveryError",
                "LatestNotificationError",
                "LatestDeliveryTime",
                "LatestNotificationTime",
                "StartLoggingTime",
                "StopLoggingTime",
                "LatestCloudWatchLogsDeliveryError",
                "LatestCloudWatchLogsDeliveryTime",
                "LatestDigestDeliveryTime",
                "LatestDigestDeliveryError",
                "LatestDeliveryAttemptTime",
                "LatestNotificationAttemptTime",
                "LatestNotificationAttemptSucceeded",
                "LatestDeliveryAttemptSucceeded",
                "TimeLoggingStarted",
                "TimeLoggingStopped",
            )
            return {"trail": {k: trail.get(k) for k in keys}}
        else:
            return {"trail": None}
    except ClientError as e:
        err = __utils__["boto3.get_error"](e)
        if e.response.get("Error", {}).get("Code") == "TrailNotFoundException":
            return {"trail": None}
        return {"error": __utils__["boto3.get_error"](e)}


def list(region=None, key=None, keyid=None, profile=None):
    """
    List all trails

    Returns list of trails

    CLI Example:

    .. code-block:: yaml

        policies:
          - {...}
          - {...}

    """
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        trails = conn.describe_trails()
        if not bool(trails.get("trailList")):
            log.warning("No trails found")
        return {"trails": trails.get("trailList", [])}
    except ClientError as e:
        return {"error": __utils__["boto3.get_error"](e)}


def update(
    Name,
    S3BucketName,
    S3KeyPrefix=None,
    SnsTopicName=None,
    IncludeGlobalServiceEvents=None,
    IsMultiRegionTrail=None,
    EnableLogFileValidation=None,
    CloudWatchLogsLogGroupArn=None,
    CloudWatchLogsRoleArn=None,
    KmsKeyId=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Given a valid config, update a trail.

    Returns {created: true} if the trail was created and returns
    {created: False} if the trail was not created.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_cloudtrail.update my_trail my_bucket

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        kwargs = {}
        for arg in (
            "S3KeyPrefix",
            "SnsTopicName",
            "IncludeGlobalServiceEvents",
            "IsMultiRegionTrail",
            "EnableLogFileValidation",
            "CloudWatchLogsLogGroupArn",
            "CloudWatchLogsRoleArn",
            "KmsKeyId",
        ):
            if locals()[arg] is not None:
                kwargs[arg] = locals()[arg]
        trail = conn.update_trail(Name=Name, S3BucketName=S3BucketName, **kwargs)
        if trail:
            log.info("The updated trail name is %s", trail["Name"])

            return {"updated": True, "name": trail["Name"]}
        else:
            log.warning("Trail was not created")
            return {"updated": False}
    except ClientError as e:
        return {"updated": False, "error": __utils__["boto3.get_error"](e)}


def start_logging(Name, region=None, key=None, keyid=None, profile=None):
    """
    Start logging for a trail

    Returns {started: true} if the trail was started and returns
    {started: False} if the trail was not started.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_cloudtrail.start_logging my_trail

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        conn.start_logging(Name=Name)
        return {"started": True}
    except ClientError as e:
        return {"started": False, "error": __utils__["boto3.get_error"](e)}


def stop_logging(Name, region=None, key=None, keyid=None, profile=None):
    """
    Stop logging for a trail

    Returns {stopped: true} if the trail was stopped and returns
    {stopped: False} if the trail was not stopped.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_cloudtrail.stop_logging my_trail

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        conn.stop_logging(Name=Name)
        return {"stopped": True}
    except ClientError as e:
        return {"stopped": False, "error": __utils__["boto3.get_error"](e)}


def _get_trail_arn(name, region=None, key=None, keyid=None, profile=None):
    if name.startswith("arn:aws:cloudtrail:"):
        return name

    account_id = __salt__["boto_iam.get_account_id"](
        region=region, key=key, keyid=keyid, profile=profile
    )
    if profile and "region" in profile:
        region = profile["region"]
    if region is None:
        region = "us-east-1"
    return "arn:aws:cloudtrail:{}:{}:trail/{}".format(region, account_id, name)


def add_tags(Name, region=None, key=None, keyid=None, profile=None, **kwargs):
    """
    Add tags to a trail

    Returns {tagged: true} if the trail was tagged and returns
    {tagged: False} if the trail was not tagged.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_cloudtrail.add_tags my_trail tag_a=tag_value tag_b=tag_value

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        tagslist = []
        for k, v in kwargs.items():
            if str(k).startswith("__"):
                continue
            tagslist.append({"Key": str(k), "Value": str(v)})
        conn.add_tags(
            ResourceId=_get_trail_arn(
                Name, region=region, key=key, keyid=keyid, profile=profile
            ),
            TagsList=tagslist,
        )
        return {"tagged": True}
    except ClientError as e:
        return {"tagged": False, "error": __utils__["boto3.get_error"](e)}


def remove_tags(Name, region=None, key=None, keyid=None, profile=None, **kwargs):
    """
    Remove tags from a trail

    Returns {tagged: true} if the trail was tagged and returns
    {tagged: False} if the trail was not tagged.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_cloudtrail.remove_tags my_trail tag_a=tag_value tag_b=tag_value

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        tagslist = []
        for k, v in kwargs.items():
            if str(k).startswith("__"):
                continue
            tagslist.append({"Key": str(k), "Value": str(v)})
        conn.remove_tags(
            ResourceId=_get_trail_arn(
                Name, region=region, key=key, keyid=keyid, profile=profile
            ),
            TagsList=tagslist,
        )
        return {"tagged": True}
    except ClientError as e:
        return {"tagged": False, "error": __utils__["boto3.get_error"](e)}


def list_tags(Name, region=None, key=None, keyid=None, profile=None):
    """
    List tags of a trail

    Returns:
        tags:
          - {...}
          - {...}

    CLI Example:

    .. code-block:: bash

        salt myminion boto_cloudtrail.list_tags my_trail

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        rid = _get_trail_arn(Name, region=region, key=key, keyid=keyid, profile=profile)
        ret = conn.list_tags(ResourceIdList=[rid])
        tlist = ret.get("ResourceTagList", []).pop().get("TagsList")
        tagdict = {}
        for tag in tlist:
            tagdict[tag.get("Key")] = tag.get("Value")
        return {"tags": tagdict}
    except ClientError as e:
        return {"error": __utils__["boto3.get_error"](e)}



"""
A Salt Util for working with the Registry.pol file. The Registry.pol file is the
source of truth for registry settings that are configured via LGPO.
"""
import logging
import os
import re
import struct

import salt.modules.win_file
import salt.utils.files
import salt.utils.win_reg
from salt.exceptions import CommandExecutionError

CLASS_INFO = {
    "User": {
        "policy_path": os.path.join(
            os.getenv("WINDIR", r"C:\Windows"),
            "System32",
            "GroupPolicy",
            "User",
            "Registry.pol",
        ),
        "hive": "HKEY_USERS",
        "lgpo_section": "User Configuration",
        "gpt_extension_location": "gPCUserExtensionNames",
        "gpt_extension_guid": "[{35378EAC-683F-11D2-A89A-00C04FBBCFA2}{D02B1F73-3407-48AE-BA88-E8213C6761F1}]",
    },
    "Machine": {
        "policy_path": os.path.join(
            os.getenv("WINDIR", r"C:\Windows"),
            "System32",
            "GroupPolicy",
            "Machine",
            "Registry.pol",
        ),
        "hive": "HKEY_LOCAL_MACHINE",
        "lgpo_section": "Computer Configuration",
        "gpt_extension_location": "gPCMachineExtensionNames",
        "gpt_extension_guid": "[{35378EAC-683F-11D2-A89A-00C04FBBCFA2}{D02B1F72-3407-48AE-BA88-E8213C6761F1}]",
    },
}
REG_POL_HEADER = "\u5250\u6765\x01\x00"
GPT_INI_PATH = os.path.join(
    os.getenv("WINDIR", "C:\\Windows"), "System32", "GroupPolicy", "gpt.ini"
)

log = logging.getLogger(__name__)

__virtualname__ = "lgpo_reg"


def __virtual__():
    """
    Only works on Windows with the lgpo_reg module
    """
    if not salt.utils.platform.is_windows():
        return False, "LGPO_REG Util: Only available on Windows"

    return __virtualname__


def search_reg_pol(search_string, policy_data):
    """
    Helper function to do a regex search of a string value in policy_data.
    This is used to search the policy data from a registry.pol file or from
    gpt.ini

    Args:
        search_string (str): The string to search for

        policy_data (str): The data to be searched

    Returns:
        bool: ``True`` if the regex search_string is found, otherwise ``False``
    """
    if policy_data:
        if search_string:
            log.debug("LGPO_REG Util: Searching for %s", search_string)
            match = re.search(search_string, policy_data, re.IGNORECASE)
            if match:
                log.debug("LGPO_REG Util: Found")
                return True
    return False


def read_reg_pol_file(reg_pol_path):
    """
    Helper function to read the content of the Registry.pol file

    Args:
        reg_pol_path (str): The path to the Registry.pol file

    Returns:
        bytes: The data as contained in the Registry.pol file
    """
    return_data = None
    if os.path.exists(reg_pol_path):
        log.debug("LGPO_REG Util: Reading from %s", reg_pol_path)
        with salt.utils.files.fopen(reg_pol_path, "rb") as pol_file:
            return_data = pol_file.read()
    return return_data


def write_reg_pol_data(
    data_to_write,
    policy_file_path,
    gpt_extension,
    gpt_extension_guid,
    gpt_ini_path=GPT_INI_PATH,
):
    """
    Helper function to actually write the data to a Registry.pol file

    Also updates/edits the gpt.ini file to include the ADM policy extensions
    to let the computer know user and/or machine registry policy files need
    to be processed

    Args:
        data_to_write (bytes): Data to write into the user/machine registry.pol
            file

        policy_file_path (str): Path to the registry.pol file

        gpt_extension (str): GPT extension list name from _policy_info class for
            this registry class gpt_extension_location

        gpt_extension_guid (str): ADMX registry extension guid for the class

        gpt_ini_path (str): The path to the gpt.ini file

    Returns:
        bool: True if successful

    Raises:
        CommandExecutionError: On failure
    """
    # Write Registry.pol file
    if not os.path.exists(policy_file_path):
        log.debug("LGPO_REG Util: Creating parent directories for Registry.pol")
        salt.modules.win_file.makedirs_(policy_file_path)
    if data_to_write is None:
        data_to_write = b""
    try:
        with salt.utils.files.fopen(policy_file_path, "wb") as pol_file:
            reg_pol_header = REG_POL_HEADER.encode("utf-16-le")
            if not data_to_write.startswith(reg_pol_header):
                log.debug("LGPO_REG Util: Writing header to %s", policy_file_path)
                pol_file.write(reg_pol_header)
            log.debug("LGPO_REG Util: Writing to %s", policy_file_path)
            pol_file.write(data_to_write)
    # TODO: This needs to be more specific
    except Exception as e:  # pylint: disable=broad-except
        msg = (
            "An error occurred attempting to write to {}, the exception was: {}".format(
                policy_file_path, e
            )
        )
        log.exception(msg)
        raise CommandExecutionError(msg)

    # Write the gpt.ini file
    gpt_ini_data = ""
    if os.path.exists(gpt_ini_path):
        with salt.utils.files.fopen(gpt_ini_path, "r") as gpt_file:
            gpt_ini_data = gpt_file.read()
        # Make sure it has Windows Style line endings
        gpt_ini_data = (
            gpt_ini_data.replace("\r\n", "_|-")
            .replace("\n", "_|-")
            .replace("_|-", "\r\n")
        )
    if not search_reg_pol(r"\[General\]\r\n", gpt_ini_data):
        log.debug("LGPO_REG Util: Adding [General] section to gpt.ini")
        gpt_ini_data = "[General]\r\n" + gpt_ini_data
    if search_reg_pol(r"{}=".format(re.escape(gpt_extension)), gpt_ini_data):
        # ensure the line contains the ADM guid
        gpt_ext_loc = re.search(
            r"^{}=.*\r\n".format(re.escape(gpt_extension)),
            gpt_ini_data,
            re.IGNORECASE | re.MULTILINE,
        )
        gpt_ext_str = gpt_ini_data[gpt_ext_loc.start() : gpt_ext_loc.end()]
        if not search_reg_pol(
            search_string=r"{}".format(re.escape(gpt_extension_guid)),
            policy_data=gpt_ext_str,
        ):
            log.debug("LGPO_REG Util: Inserting gpt extension GUID")
            gpt_ext_str = gpt_ext_str.split("=")
            gpt_ext_str[1] = gpt_extension_guid + gpt_ext_str[1]
            gpt_ext_str = "=".join(gpt_ext_str)
            gpt_ini_data = (
                gpt_ini_data[0 : gpt_ext_loc.start()]
                + gpt_ext_str
                + gpt_ini_data[gpt_ext_loc.end() :]
            )
    else:
        general_location = re.search(
            r"^\[General\]\r\n", gpt_ini_data, re.IGNORECASE | re.MULTILINE
        )
        gpt_ini_data = "{}{}={}\r\n{}".format(
            gpt_ini_data[general_location.start() : general_location.end()],
            gpt_extension,
            gpt_extension_guid,
            gpt_ini_data[general_location.end() :],
        )
    # https://technet.microsoft.com/en-us/library/cc978247.aspx
    if search_reg_pol(r"Version=", gpt_ini_data):
        version_loc = re.search(
            r"^Version=.*\r\n", gpt_ini_data, re.IGNORECASE | re.MULTILINE
        )
        version_str = gpt_ini_data[version_loc.start() : version_loc.end()]
        version_str = version_str.split("=")
        version_nums = struct.unpack(b">2H", struct.pack(b">I", int(version_str[1])))
        if gpt_extension.lower() == "gPCMachineExtensionNames".lower():
            version_nums = (version_nums[0], version_nums[1] + 1)
        elif gpt_extension.lower() == "gPCUserExtensionNames".lower():
            version_nums = (version_nums[0] + 1, version_nums[1])
        version_num = struct.unpack(b">I", struct.pack(b">2H", *version_nums))[0]
        gpt_ini_data = "{}{}={}\r\n{}".format(
            gpt_ini_data[0 : version_loc.start()],
            "Version",
            version_num,
            gpt_ini_data[version_loc.end() :],
        )
    else:
        general_location = re.search(
            r"^\[General\]\r\n", gpt_ini_data, re.IGNORECASE | re.MULTILINE
        )
        if gpt_extension.lower() == "gPCMachineExtensionNames".lower():
            version_nums = (0, 1)
        elif gpt_extension.lower() == "gPCUserExtensionNames".lower():
            version_nums = (1, 0)
        gpt_ini_data = "{}{}={}\r\n{}".format(
            gpt_ini_data[general_location.start() : general_location.end()],
            "Version",
            int(
                "{}{}".format(
                    str(version_nums[0]).zfill(4),
                    str(version_nums[1]).zfill(4),
                ),
                16,
            ),
            gpt_ini_data[general_location.end() :],
        )
    if gpt_ini_data:
        try:
            with salt.utils.files.fopen(gpt_ini_path, "w") as gpt_file:
                gpt_file.write(gpt_ini_data)
        # TODO: This needs to be more specific
        except Exception as e:  # pylint: disable=broad-except
            msg = (
                "An error occurred attempting to write the gpg.ini file.\n"
                "path: {}\n"
                "exception: {}".format(gpt_ini_path, e)
            )
            log.exception(msg)
            raise CommandExecutionError(msg)
    return True


def reg_pol_to_dict(policy_data):
    """
    Convert the data obtained from a Registry.pol file to a dictionary.

    Args:
        policy_data (bytes): The data as retrieved from the Registry.pol file

    Raises:
        SaltInvocationError: Invalid or corrupt policy data

    Returns:
        dict: A dictionary representation of the Registry.pol data
    """
    # https://learn.microsoft.com/en-us/previous-versions/windows/desktop/Policy/registry-policy-file-format
    # https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-gpreg/5c092c22-bf6b-4e7f-b180-b20743d368f5

    reg_pol_header = REG_POL_HEADER.encode("utf-16-le")

    # If policy_data is None, that means the Registry.pol file is missing
    # So, we'll create it
    if policy_data is None:
        policy_data = reg_pol_header

    if not policy_data.startswith(reg_pol_header):
        msg = "LGPO_REG Util: Invalid Header. Registry.pol may be corrupt"
        raise CommandExecutionError(msg)

    # Strip the header information, we don't care about it now
    pol_file_data = policy_data.lstrip(reg_pol_header)

    if not pol_file_data:
        log.debug("LGPO_REG Util: No registry.pol data to return")
        return {}

    def strip_field_end(value):
        while value[-2:] == b"\x00\x00":
            value = value[:-2]
        return value

    log.debug("LGPO_REG Util: Unpacking reg pol data")
    reg_pol = {}
    # Each policy is inside square braces. In this case they're encoded
    # utf-16-le, so there's a null-byte for each character
    for policy in pol_file_data.split(b"]\x00[\x00"):
        # Now remove the lingering square braces
        policy = policy.replace(b"]\x00", b"").replace(b"[\x00", b"")
        # Each policy element is delimited by a semi-colon, encoded utf-16-le
        # The first 4 are the key, name, type and size
        # all remaining is data
        key, v_name, v_type, v_size, v_data = policy.split(b";\x00", 4)
        # Removing trailing null-bytes
        key = strip_field_end(key).decode("utf-16-le")
        v_name = strip_field_end(v_name).decode("utf-16-le")
        # v_type is one of (0, 1, 2, 3, 4, 5, 7, 11) as 32-bit little-endian
        v_type = struct.unpack("<i", v_type)[0]
        if v_type == 0:
            # REG_NONE : No Type
            # We don't know what this data is, so don't do anything
            pass
        elif v_type in (1, 2):
            # REG_SZ : String Type
            # REG_EXPAND_SZ : String with Environment Variables, ie %PATH%
            v_data = strip_field_end(v_data).decode("utf-16-le")
        elif v_type == 4:
            # REG_DWORD : 32-bit little endian
            v_data = struct.unpack("<i", v_data)[0]
        elif v_type == 5:
            # REG_DWORD : 32-bit big endian
            v_data = struct.unpack(">i", v_data)[0]
        elif v_type == 7:
            # REG_MULTI_SZ : Multiple strings, delimited by \x00
            v_data = strip_field_end(v_data)
            if not v_data:
                v_data = None
            else:
                v_data = v_data.decode("utf-16-le").split("\x00")
        elif v_type == 11:
            # REG_QWORD : 64-bit little endian
            v_data = struct.unpack("<q", v_data)[0]
        else:
            msg = "LGPO_REG Util: Found unknown registry type: {}".format(v_type)
            raise CommandExecutionError(msg)

        # Lookup the REG Type from the number
        reg = salt.utils.win_reg.Registry()
        v_type = reg.vtype_reverse.get(v_type, "REG_NONE")

        # Make the dictionary entries
        reg_pol.setdefault(key, {})
        if not v_name:
            reg_pol[key]["*"] = "CREATEKEY"
        else:
            reg_pol[key][v_name] = {"type": v_type, "data": v_data}

    return reg_pol


def dict_to_reg_pol(data):
    """
    Convert a dictionary to the bytes format expected by the Registry.pol file

    Args:
        data (dict): A dictionary containing the contents to be converted

    Returns:
        bytes: The data to be written to the Registry.pol file
    """
    # https://learn.microsoft.com/en-us/previous-versions/windows/desktop/Policy/registry-policy-file-format
    # https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-gpreg/5c092c22-bf6b-4e7f-b180-b20743d368f5
    reg = salt.utils.win_reg.Registry()

    pol_enter_delim = "[".encode("utf-16-le")
    pol_exit_delim = "]".encode("utf-16-le")
    pol_section_delim = ";".encode("utf-16-le")
    pol_section_term = "\x00".encode("utf-16-le")

    policies = []
    for key, value in data.items():
        for v_name, d in value.items():
            # Handle CREATEKEY entry
            if v_name == "*" and d == "CREATEKEY":
                v_name = ""
                d = {"type": "REG_NONE"}

            try:
                v_type = reg.vtype[d["type"]]
            except KeyError:
                msg = "LGPO_REG Util: Found unknown registry type: {}".format(d["type"])
                raise CommandExecutionError(msg)

            # The first three items are pretty straight forward
            policy = [
                # Key followed by null byte
                "{}".format(key).encode("utf-16-le") + pol_section_term,
                # Value name followed by null byte
                "{}".format(v_name).encode("utf-16-le") + pol_section_term,
                # Type in 32-bit little-endian
                struct.pack("<i", v_type),
            ]
            # The data is encoded depending on the Type
            if v_type == 0:
                # REG_NONE : No value
                v_data = b""
            elif v_type in (1, 2):
                # REG_SZ : String Type
                # REG_EXPAND_SZ : String with Environment Variables, ie %PATH%
                # Value followed by null byte
                v_data = d["data"].encode("utf-16-le") + pol_section_term
            elif v_type == 4:
                # REG_DWORD : Little Endian
                # 32-bit little endian
                v_data = struct.pack("<i", int(d["data"]))
            elif v_type == 5:
                # REG_DWORD : Big Endian (not common)
                # 32-bit big endian
                v_data = struct.pack(">i", int(d["data"]))
            elif v_type == 7:
                # REG_MULTI_SZ : Multiple strings
                # Each element is delimited by \x00, terminated by \x00\x00
                # Then the entire output is terminated with a null byte
                if d["data"] is None:
                    # An None value just gets the section terminator
                    v_data = pol_section_term
                elif len(d["data"]) == 0:
                    # An empty list just gets the section terminator
                    v_data = pol_section_term
                elif len(d["data"]) == 1 and not d["data"][0]:
                    # An list with an empty value just gets the section terminator
                    v_data = pol_section_term
                else:
                    # All others will be joined with a null byte, the list
                    # terminated, and the entire section terminated
                    v_data = (
                        "\x00".join(d["data"]).encode("utf-16-le")
                        + pol_section_term
                        + pol_section_term
                    )
            elif v_type == 11:
                # REG_QWORD : Little Endian
                # 64-bit little endian
                v_data = struct.pack("<q", int(d["data"]))

            # Now that we have the data in the right format, let's calculate its size
            # 16-bit little endian with null terminator
            if len(v_data) > 65535:
                msg = "LGPO_REG Util: Size exceeds 65535 bytes"
                raise CommandExecutionError(msg)
            v_size = len(v_data).to_bytes(2, "little") + pol_section_term

            policy.append(v_size)
            policy.append(v_data)

            policies.append(pol_section_delim.join(policy))

    policy_file_data = REG_POL_HEADER.encode("utf-16-le")
    for policy in policies:
        policy_file_data += pol_enter_delim + policy + pol_exit_delim

    return policy_file_data



import pathlib

import attr
import pytest

import salt.transport.ipc
import salt.utils.platform
from salt.ext.tornado import locks

pytestmark = [
    # Windows does not support POSIX IPC
    pytest.mark.skip_on_windows,
]


@attr.s(frozen=True, slots=True)
class PayloadHandler:
    payloads = attr.ib(init=False, default=attr.Factory(list))

    async def handle_payload(self, payload, reply_func):
        self.payloads.append(payload)
        await reply_func(payload)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.payloads.clear()


@attr.s(frozen=True, slots=True)
class IPCTester:
    io_loop = attr.ib()
    socket_path = attr.ib()
    server = attr.ib()
    client = attr.ib()
    payloads = attr.ib(default=attr.Factory(list))
    payload_ack = attr.ib(default=attr.Factory(locks.Condition))

    @server.default
    def _server_default(self):
        return salt.transport.ipc.IPCMessageServer(
            self.socket_path,
            io_loop=self.io_loop,
            payload_handler=self.handle_payload,
        )

    @client.default
    def _client_default(self):
        return salt.transport.ipc.IPCMessageClient(
            self.socket_path,
            io_loop=self.io_loop,
        )

    async def handle_payload(self, payload, reply_func):
        self.payloads.append(payload)
        await reply_func(payload)
        self.payload_ack.notify()

    def new_client(self):
        return IPCTester(
            io_loop=self.io_loop,
            socket_path=self.socket_path,
            server=self.server,
            payloads=self.payloads,
            payload_ack=self.payload_ack,
        )

    async def send(self, payload, timeout=60):
        ret = await self.client.send(payload)
        await self.payload_ack.wait(self.io_loop.time() + timeout)
        return ret

    def __enter__(self):
        self.io_loop.add_callback(self.server.start)
        self.io_loop.add_callback(self.client.connect)
        return self

    def __exit__(self, *args):
        self.client.close()
        self.server.close()


@pytest.fixture
def ipc_socket_path(tmp_path):
    if salt.utils.platform.is_darwin():
        # A shorter path so that we don't hit the AF_UNIX path too long
        tmp_path = pathlib.Path("/tmp").resolve()
    _socket_path = tmp_path / "ipc-test.ipc"
    try:
        yield _socket_path
    finally:
        if _socket_path.exists():
            _socket_path.unlink()


@pytest.fixture
def channel(io_loop, ipc_socket_path):
    _ipc_tester = IPCTester(io_loop=io_loop, socket_path=str(ipc_socket_path))
    with _ipc_tester:
        yield _ipc_tester


async def test_basic_send(channel):
    msg = {"foo": "bar", "stop": True}
    await channel.send(msg)
    assert channel.payloads[0] == msg


async def test_send_many(channel):
    msgs = []
    for i in range(0, 1000):
        msgs.append("test_many_send_{}".format(i))

    for msg in msgs:
        await channel.send(msg)
    assert channel.payloads == msgs


async def test_very_big_message(channel):
    long_str = "".join([str(num) for num in range(10**5)])
    msg = {"long_str": long_str, "stop": True}
    await channel.send(msg)
    assert channel.payloads[0] == msg


async def test_multistream_sends(channel):
    new_channel = channel.new_client()
    with new_channel:
        assert channel.client is not new_channel.client
        await new_channel.send("foo")
        await channel.send("foo")
    assert channel.payloads == ["foo", "foo"]


async def test_multistream_error_sends(channel):
    new_channel = channel.new_client()
    with new_channel:
        assert channel.client is not new_channel.client
        await new_channel.send(None)
        await channel.send(None)
        await new_channel.send("foo")
        await channel.send("foo")
    assert channel.payloads == [None, None, "foo", "foo"]



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""

import pytest

import salt.modules.firewalld as firewalld
from tests.support.helpers import dedent
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {firewalld: {}}


def test_version():
    """
    Test for Return version from firewall-cmd
    """
    with patch.object(firewalld, "__firewall_cmd", return_value=2):
        assert firewalld.version() == 2


def test_default_zone():
    """
    Test for Print default zone for connections and interfaces
    """
    with patch.object(firewalld, "__firewall_cmd", return_value="A"):
        assert firewalld.default_zone() == "A"


def test_list_zones():
    """
    Test for List everything added for or enabled in all zones
    """
    # pylint: disable=trailing-whitespace
    firewall_cmd_ret = dedent(
        """\
            nm-shared
              target: ACCEPT
              icmp-block-inversion: no
              interfaces:
              sources:
              services: dhcp dns ssh
              ports:
              protocols: icmp ipv6-icmp
              masquerade: no
              forward-ports:
              source-ports:
              icmp-blocks:
              rich rules:
            \trule priority="32767" reject

            public
              target: default
              icmp-block-inversion: no
              interfaces:
              sources:
              services: cockpit dhcpv6-client ssh
              ports:
              protocols:
              masquerade: no
              forward-ports:
              source-ports:
              icmp-blocks:
              rich rules:
            """
    )
    # pylint: enable=trailing-whitespace
    ret = {
        "nm-shared": {
            "forward-ports": [""],
            "icmp-block-inversion": ["no"],
            "icmp-blocks": [""],
            "interfaces": [""],
            "masquerade": ["no"],
            "ports": [""],
            "protocols": ["icmp ipv6-icmp"],
            "rich rules": ["", 'rule priority="32767" reject'],
            "services": ["dhcp dns ssh"],
            "source-ports": [""],
            "sources": [""],
            "target": ["ACCEPT"],
        },
        "public": {
            "forward-ports": [""],
            "icmp-block-inversion": ["no"],
            "icmp-blocks": [""],
            "interfaces": [""],
            "masquerade": ["no"],
            "ports": [""],
            "protocols": [""],
            "rich rules": [""],
            "services": ["cockpit dhcpv6-client ssh"],
            "source-ports": [""],
            "sources": [""],
            "target": ["default"],
        },
    }

    with patch.object(firewalld, "__firewall_cmd", return_value=firewall_cmd_ret):
        assert firewalld.list_zones() == ret


def test_list_zones_empty_response():
    """
    Test list_zones if firewall-cmd call returns nothing
    """
    with patch.object(firewalld, "__firewall_cmd", return_value=""):
        assert firewalld.list_zones() == {}


def test_get_zones():
    """
    Test for Print predefined zones
    """
    with patch.object(firewalld, "__firewall_cmd", return_value="A"):
        assert firewalld.get_zones() == ["A"]


def test_get_services():
    """
    Test for Print predefined services
    """
    with patch.object(firewalld, "__firewall_cmd", return_value="A"):
        assert firewalld.get_services() == ["A"]


def test_get_icmp_types():
    """
    Test for Print predefined icmptypes
    """
    with patch.object(firewalld, "__firewall_cmd", return_value="A"):
        assert firewalld.get_icmp_types() == ["A"]


def test_new_zone():
    """
    Test for Add a new zone
    """
    with patch.object(firewalld, "__mgmt", return_value="success"):
        mock = MagicMock(return_value="A")
        with patch.object(firewalld, "__firewall_cmd", mock):
            assert firewalld.new_zone("zone") == "A"

    with patch.object(firewalld, "__mgmt", return_value="A"):
        assert firewalld.new_zone("zone") == "A"

    with patch.object(firewalld, "__mgmt", return_value="A"):
        assert firewalld.new_zone("zone", False) == "A"


def test_delete_zone():
    """
    Test for Delete an existing zone
    """
    with patch.object(firewalld, "__mgmt", return_value="success"):
        with patch.object(firewalld, "__firewall_cmd", return_value="A"):
            assert firewalld.delete_zone("zone") == "A"

    with patch.object(firewalld, "__mgmt", return_value="A"):
        assert firewalld.delete_zone("zone") == "A"

    mock = MagicMock(return_value="A")
    with patch.object(firewalld, "__mgmt", return_value="A"):
        assert firewalld.delete_zone("zone", False) == "A"


def test_set_default_zone():
    """
    Test for Set default zone
    """
    with patch.object(firewalld, "__firewall_cmd", return_value="A"):
        assert firewalld.set_default_zone("zone") == "A"


def test_new_service():
    """
    Test for Add a new service
    """
    with patch.object(firewalld, "__mgmt", return_value="success"):
        mock = MagicMock(return_value="A")
        with patch.object(firewalld, "__firewall_cmd", return_value="A"):
            assert firewalld.new_service("zone") == "A"

    with patch.object(firewalld, "__mgmt", return_value="A"):
        assert firewalld.new_service("zone") == "A"

    with patch.object(firewalld, "__mgmt", return_value="A"):
        assert firewalld.new_service("zone", False) == "A"


def test_delete_service():
    """
    Test for Delete an existing service
    """
    with patch.object(firewalld, "__mgmt", return_value="success"):
        mock = MagicMock(return_value="A")
        with patch.object(firewalld, "__firewall_cmd", return_value="A"):
            assert firewalld.delete_service("name") == "A"

    with patch.object(firewalld, "__mgmt", return_value="A"):
        assert firewalld.delete_service("name") == "A"

    with patch.object(firewalld, "__mgmt", return_value="A"):
        assert firewalld.delete_service("name", False) == "A"


def test_list_all():
    """
    Test for List everything added for or enabled in a zone
    """
    # pylint: disable=trailing-whitespace
    firewall_cmd_ret = dedent(
        """\
        public
          target: default
          icmp-block-inversion: no
          interfaces: eth0
          sources:
          services: cockpit dhcpv6-client ssh
          ports:
          protocols:
          masquerade: no
          forward-ports:
          source-ports:
          icmp-blocks:
          rich rules:
        """
    )
    # pylint: enable=trailing-whitespace
    ret = {
        "public": {
            "forward-ports": [""],
            "icmp-block-inversion": ["no"],
            "icmp-blocks": [""],
            "interfaces": ["eth0"],
            "masquerade": ["no"],
            "ports": [""],
            "protocols": [""],
            "rich rules": [""],
            "services": ["cockpit dhcpv6-client ssh"],
            "source-ports": [""],
            "sources": [""],
            "target": ["default"],
        }
    }
    with patch.object(firewalld, "__firewall_cmd", return_value=firewall_cmd_ret):
        assert firewalld.list_all() == ret


def test_list_all_empty_response():
    """
    Test list_all if firewall-cmd call returns nothing
    """
    with patch.object(firewalld, "__firewall_cmd", return_value=""):
        assert firewalld.list_all() == {}


def test_list_services():
    """
    Test for List services added for zone as a space separated list.
    """
    with patch.object(firewalld, "__firewall_cmd", return_value=""):
        assert firewalld.list_services() == []


def test_add_service():
    """
    Test for Add a service for zone
    """
    with patch.object(firewalld, "__firewall_cmd", return_value=""):
        assert firewalld.add_service("name") == ""


def test_remove_service():
    """
    Test for Remove a service from zone
    """
    with patch.object(firewalld, "__firewall_cmd", return_value=""):
        assert firewalld.remove_service("name") == ""


def test_add_masquerade():
    """
    Test for adding masquerade
    """
    with patch.object(firewalld, "__firewall_cmd", return_value="success"):
        assert firewalld.add_masquerade("name") == "success"


def test_remove_masquerade():
    """
    Test for removing masquerade
    """
    with patch.object(firewalld, "__firewall_cmd", return_value="success"):
        assert firewalld.remove_masquerade("name") == "success"


def test_add_port():
    """
    Test adding a port to a specific zone
    """
    with patch.object(firewalld, "__firewall_cmd", return_value="success"):
        assert firewalld.add_port("zone", "80/tcp") == "success"


def test_remove_port():
    """
    Test removing a port from a specific zone
    """
    with patch.object(firewalld, "__firewall_cmd", return_value="success"):
        assert firewalld.remove_port("zone", "80/tcp") == "success"


def test_list_ports():
    """
    Test listing ports within a zone
    """
    ret = "22/tcp 53/udp 53/tcp"
    exp = ["22/tcp", "53/udp", "53/tcp"]

    with patch.object(firewalld, "__firewall_cmd", return_value=ret):
        assert firewalld.list_ports("zone") == exp


def test_add_port_fwd():
    """
    Test adding port forwarding on a zone
    """
    with patch.object(firewalld, "__firewall_cmd", return_value="success"):
        assert firewalld.add_port_fwd("zone", "22", "2222", "tcp") == "success"


def test_remove_port_fwd():
    """
    Test removing port forwarding on a zone
    """
    with patch.object(firewalld, "__firewall_cmd", return_value="success"):
        assert firewalld.remove_port_fwd("zone", "22", "2222", "tcp") == "success"


def test_list_port_fwd():
    """
    Test listing all port forwarding for a zone
    """
    ret = "port=23:proto=tcp:toport=8080:toaddr=\nport=80:proto=tcp:toport=443:toaddr="
    exp = [
        {
            "Destination address": "",
            "Destination port": "8080",
            "Protocol": "tcp",
            "Source port": "23",
        },
        {
            "Destination address": "",
            "Destination port": "443",
            "Protocol": "tcp",
            "Source port": "80",
        },
    ]

    with patch.object(firewalld, "__firewall_cmd", return_value=ret):
        assert firewalld.list_port_fwd("zone") == exp


def test_block_icmp():
    """
    Test ICMP block
    """
    with patch.object(firewalld, "__firewall_cmd", return_value="success"):
        with patch.object(firewalld, "get_icmp_types", return_value="echo-reply"):
            assert firewalld.block_icmp("zone", "echo-reply") == "success"

    with patch.object(firewalld, "__firewall_cmd"):
        assert not firewalld.block_icmp("zone", "echo-reply")


def test_allow_icmp():
    """
    Test ICMP allow
    """
    with patch.object(firewalld, "__firewall_cmd", return_value="success"):
        with patch.object(firewalld, "get_icmp_types", return_value="echo-reply"):
            assert firewalld.allow_icmp("zone", "echo-reply") == "success"

    with patch.object(firewalld, "__firewall_cmd", return_value="success"):
        assert not firewalld.allow_icmp("zone", "echo-reply")


def test_list_icmp_block():
    """
    Test ICMP block list
    """
    ret = "echo-reply echo-request"
    exp = ["echo-reply", "echo-request"]

    with patch.object(firewalld, "__firewall_cmd", return_value=ret):
        assert firewalld.list_icmp_block("zone") == exp


def test_get_rich_rules():
    """
    Test listing rich rules bound to a zone
    """
    with patch.object(firewalld, "__firewall_cmd", return_value=""):
        assert firewalld.get_rich_rules("zone") == []


def test_add_rich_rule():
    """
    Test adding a rich rule to a zone
    """
    with patch.object(firewalld, "__firewall_cmd", return_value="success"):
        assert (
            firewalld.add_rich_rule(
                "zone", 'rule family="ipv4" source address="1.2.3.4" accept'
            )
            == "success"
        )


def test_remove_rich_rule():
    """
    Test removing a rich rule to a zone
    """
    with patch.object(firewalld, "__firewall_cmd", return_value="success"):
        assert (
            firewalld.remove_rich_rule(
                "zone", 'rule family="ipv4" source address="1.2.3.4" accept'
            )
            == "success"
        )



import re

import pytest

import salt.modules.win_file as win_file
from salt.exceptions import CommandExecutionError

pytestmark = [pytest.mark.windows_whitelisted, pytest.mark.skip_unless_on_windows]


def test__get_version_os():
    expected = ["32-bit Windows", "Windows NT"]
    result = win_file._get_version_os(0x00040004)
    assert result == expected


def test__get_version_type_application():
    expected = "Application"
    result = win_file._get_version_type(1, 0)
    assert result == expected


def test__get_version_type_driver():
    expected = "Printer Driver"
    result = win_file._get_version_type(3, 1)
    assert result == expected


def test__get_version_type_font():
    expected = "TrueType Font"
    result = win_file._get_version_type(4, 3)
    assert result == expected


def test__get_version_type_virtual_device():
    expected = "Virtual Device: 12345"
    result = win_file._get_version_type(5, 12345)
    assert result == expected


def test__get_version_exe():
    result = win_file._get_version(r"C:\Windows\notepad.exe")
    regex = re.compile(r"\d+.\d+.\d+.\d+")
    assert regex.search(result)


def test__get_version_dll():
    result = win_file._get_version(r"C:\Windows\System32\FirewallAPI.dll")
    regex = re.compile(r"\d+.\d+.\d+.\d+")
    assert regex.search(result)


def test__get_version_sys():
    result = win_file._get_version(r"C:\Windows\System32\drivers\netio.sys")
    regex = re.compile(r"\d+.\d+.\d+.\d+")
    assert regex.search(result)


def test_version_missing_file():
    with pytest.raises(CommandExecutionError):
        win_file.version("C:\\Windows\\bogus.exe")


def test_version_missing_directory():
    with pytest.raises(CommandExecutionError):
        win_file.version("C:\\Windows\\System32")


def test_version_details_missing_file():
    with pytest.raises(CommandExecutionError):
        win_file.version_details("C:\\Windows\\bogus.exe")


def test_version_details_missing_directory():
    with pytest.raises(CommandExecutionError):
        win_file.version_details("C:\\Windows\\System32")


def test_version_details_exe():
    result = win_file.version_details(r"C:\Windows\notepad.exe")
    assert result["FileDescription"] == "Notepad"
    assert result["FileType"] == "Application"
    assert result["OperatingSystem"] == ["32-bit Windows", "Windows NT"]
    regex = re.compile(r"\d+.\d+.\d+.\d+")
    assert regex.search(result["ProductVersion"])
    assert regex.search(result["Version"])


def test_version_details_dll():
    result = win_file.version_details(r"C:\Windows\System32\FirewallAPI.dll")
    assert "Firewall API" in result["FileDescription"]
    assert result["FileType"] == "DLL"
    assert result["OperatingSystem"] == ["32-bit Windows", "Windows NT"]
    regex = re.compile(r"\d+.\d+.\d+.\d+")
    assert regex.search(result["ProductVersion"])
    assert regex.search(result["Version"])


def test_version_details_sys():
    result = win_file.version_details(r"C:\Windows\System32\drivers\netio.sys")
    assert result["FileDescription"] == "Network I/O Subsystem"
    assert result["FileType"] == "Network Driver"
    assert result["OperatingSystem"] == ["32-bit Windows", "Windows NT"]
    regex = re.compile(r"\d+.\d+.\d+.\d+")
    assert regex.search(result["ProductVersion"])
    assert regex.search(result["Version"])
